{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "75c9992a-8f78-4e17-b8a7-9c64835dea2a",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "source": [
    "# First-order properties\n",
    "\n",
    "In Kohn–Sham DFT the accessible quantity is the [one-particle density](https://kthpanor.github.io/echem/docs/elec_struct/reduced_density.html#one-particle-density) that here is expressed in terms of the atomic orbitals and their associated density matrix\n",
    "\n",
    "$$\n",
    "n(\\mathbf{r}) =\n",
    "\\sum_{\\alpha, \\beta} D_{\\alpha\\beta} \\chi_\\alpha^\\ast(\\mathbf{r}) \\chi_\\beta(\\mathbf{r})\n",
    "$$\n",
    "\n",
    "Let us determine this density matrix for water at the level of B3LYP/cc-pVDZ."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "d2e9ee83-89b9-4a26-9615-fb1fdcb8dd94",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": [
     "remove-output"
    ]
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "* Warning * Environment variable OMP_NUM_THREADS not set.\n",
      "* Warning * Setting OMP_NUM_THREADS to 10.\n"
     ]
    }
   ],
   "source": [
    "import numpy as np\n",
    "import veloxchem as vlx"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "8b1b9df5-94f7-4df9-add6-439050663fc5",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "source": [
    "Define the system."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "id": "088986cd-1278-4b06-b801-ba13391e4f21",
   "metadata": {
    "editable": true,
    "scrolled": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": [
     "remove-output"
    ]
   },
   "outputs": [],
   "source": [
    "h2o_xyz = \"\"\"3\n",
    "\n",
    "O    0.000000000000        0.000000000000        0.000000000000\n",
    "H    0.000000000000        0.740848095288        0.582094932012\n",
    "H    0.000000000000       -0.740848095288        0.582094932012\n",
    "\"\"\"\n",
    "molecule = vlx.Molecule.read_xyz_string(h2o_xyz)\n",
    "basis = vlx.MolecularBasis.read(molecule, \"cc-pvdz\", ostream=None)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "id": "8a4f4fef-1107-47aa-8805-510de51d9118",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "outputs": [
    {
     "data": {
      "application/3dmoljs_load.v0": "<div id=\"3dmolviewer_17057781142609549\"  style=\"position: relative; width: 400px; height: 300px;\">\n        <p id=\"3dmolwarning_17057781142609549\" style=\"background-color:#ffcccc;color:black\">You appear to be running in JupyterLab (or JavaScript failed to load for some other reason).  You need to install the 3dmol extension: <br>\n        <tt>jupyter labextension install jupyterlab_3dmol</tt></p>\n        </div>\n<script>\n\nvar loadScriptAsync = function(uri){\n  return new Promise((resolve, reject) => {\n    //this is to ignore the existence of requirejs amd\n    var savedexports, savedmodule;\n    if (typeof exports !== 'undefined') savedexports = exports;\n    else exports = {}\n    if (typeof module !== 'undefined') savedmodule = module;\n    else module = {}\n\n    var tag = document.createElement('script');\n    tag.src = uri;\n    tag.async = true;\n    tag.onload = () => {\n        exports = savedexports;\n        module = savedmodule;\n        resolve();\n    };\n  var firstScriptTag = document.getElementsByTagName('script')[0];\n  firstScriptTag.parentNode.insertBefore(tag, firstScriptTag);\n});\n};\n\nif(typeof $3Dmolpromise === 'undefined') {\n$3Dmolpromise = null;\n  $3Dmolpromise = loadScriptAsync('https://cdnjs.cloudflare.com/ajax/libs/3Dmol/2.0.4/3Dmol-min.js');\n}\n\nvar viewer_17057781142609549 = null;\nvar warn = document.getElementById(\"3dmolwarning_17057781142609549\");\nif(warn) {\n    warn.parentNode.removeChild(warn);\n}\n$3Dmolpromise.then(function() {\nviewer_17057781142609549 = $3Dmol.createViewer(document.getElementById(\"3dmolviewer_17057781142609549\"),{backgroundColor:\"white\"});\nviewer_17057781142609549.zoomTo();\n\tviewer_17057781142609549.addModel(\"3\\n\\nO              0.000000000000         0.000000000000         0.000000000000\\nH              0.000000000000         0.740848095288         0.582094932012\\nH              0.000000000000        -0.740848095288         0.582094932012\\n\");\n\tviewer_17057781142609549.setViewStyle({\"style\": \"outline\", \"width\": 0.05});\n\tviewer_17057781142609549.setStyle({\"stick\": {}, \"sphere\": {\"scale\": 0.25}});\n\tviewer_17057781142609549.zoomTo();\nviewer_17057781142609549.render();\n});\n</script>",
      "text/html": [
       "<div id=\"3dmolviewer_17057781142609549\"  style=\"position: relative; width: 400px; height: 300px;\">\n",
       "        <p id=\"3dmolwarning_17057781142609549\" style=\"background-color:#ffcccc;color:black\">You appear to be running in JupyterLab (or JavaScript failed to load for some other reason).  You need to install the 3dmol extension: <br>\n",
       "        <tt>jupyter labextension install jupyterlab_3dmol</tt></p>\n",
       "        </div>\n",
       "<script>\n",
       "\n",
       "var loadScriptAsync = function(uri){\n",
       "  return new Promise((resolve, reject) => {\n",
       "    //this is to ignore the existence of requirejs amd\n",
       "    var savedexports, savedmodule;\n",
       "    if (typeof exports !== 'undefined') savedexports = exports;\n",
       "    else exports = {}\n",
       "    if (typeof module !== 'undefined') savedmodule = module;\n",
       "    else module = {}\n",
       "\n",
       "    var tag = document.createElement('script');\n",
       "    tag.src = uri;\n",
       "    tag.async = true;\n",
       "    tag.onload = () => {\n",
       "        exports = savedexports;\n",
       "        module = savedmodule;\n",
       "        resolve();\n",
       "    };\n",
       "  var firstScriptTag = document.getElementsByTagName('script')[0];\n",
       "  firstScriptTag.parentNode.insertBefore(tag, firstScriptTag);\n",
       "});\n",
       "};\n",
       "\n",
       "if(typeof $3Dmolpromise === 'undefined') {\n",
       "$3Dmolpromise = null;\n",
       "  $3Dmolpromise = loadScriptAsync('https://cdnjs.cloudflare.com/ajax/libs/3Dmol/2.0.4/3Dmol-min.js');\n",
       "}\n",
       "\n",
       "var viewer_17057781142609549 = null;\n",
       "var warn = document.getElementById(\"3dmolwarning_17057781142609549\");\n",
       "if(warn) {\n",
       "    warn.parentNode.removeChild(warn);\n",
       "}\n",
       "$3Dmolpromise.then(function() {\n",
       "viewer_17057781142609549 = $3Dmol.createViewer(document.getElementById(\"3dmolviewer_17057781142609549\"),{backgroundColor:\"white\"});\n",
       "viewer_17057781142609549.zoomTo();\n",
       "\tviewer_17057781142609549.addModel(\"3\\n\\nO              0.000000000000         0.000000000000         0.000000000000\\nH              0.000000000000         0.740848095288         0.582094932012\\nH              0.000000000000        -0.740848095288         0.582094932012\\n\");\n",
       "\tviewer_17057781142609549.setViewStyle({\"style\": \"outline\", \"width\": 0.05});\n",
       "\tviewer_17057781142609549.setStyle({\"stick\": {}, \"sphere\": {\"scale\": 0.25}});\n",
       "\tviewer_17057781142609549.zoomTo();\n",
       "viewer_17057781142609549.render();\n",
       "});\n",
       "</script>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "molecule.show()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "ec8b949f-4c81-43ad-9d7e-3f555652d61b",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "source": [
    "Define the method and perform an SCF optimization."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "id": "d156b1bc-882b-4afe-98c2-ca56f6917baa",
   "metadata": {
    "editable": true,
    "scrolled": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "outputs": [],
   "source": [
    "scf_drv = vlx.ScfRestrictedDriver()\n",
    "scf_drv.ostream.mute()\n",
    "scf_drv.xcfun = \"b3lyp\"\n",
    "scf_results = scf_drv.compute(molecule, basis)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "id": "1750d837-516f-4769-a386-3d1fd2e8292d",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "outputs": [],
   "source": [
    "D = scf_results[\"D_alpha\"] + scf_results[\"D_beta\"]"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "99c67afa-66d1-4b2c-99d5-6d65385bcc64",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "source": [
    "An observable associated with a scalar [one-electron operator](https://kthpanor.github.io/echem/docs/elec_struct/operators.html#one-electron-operators) in the coordinate basis \n",
    "\n",
    "$$\n",
    "\\hat{\\Omega} = \\sum_{i=1}^N \\omega(\\mathbf{r}_i)\n",
    "$$\n",
    "\n",
    "takes the form of an expectation value\n",
    "\n",
    "\\begin{align*}\n",
    "\\langle \\hat{\\Omega} \\rangle & =\n",
    "\\int \\omega(\\mathbf{r}) \\, n(\\mathbf{r}) \\,\n",
    "\\mathrm{d}^3 \\mathbf{r} \n",
    "\\end{align*}\n",
    "\n",
    "We will determine the permanent dipole moment along the $z$-axis such that \n",
    "\n",
    "$$\n",
    "\\hat{\\Omega} = \\hat{\\mu}_z ; \\qquad\n",
    "\\omega(\\mathbf{r}) = - e z\n",
    "$$"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "af56cefe-cbf2-4052-a79d-a03d7a5828e7",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "source": [
    "## Numerical integration\n",
    "\n",
    "In DFT, integrals involving the exchange–correlation kernel are evaluated [numerically](https://kthpanor.github.io/echem/docs/elec_struct/kernel_int.html). The same technique can, however, be used to evaluate first-order properties \n",
    "\n",
    "$$\n",
    "\\langle \\hat{\\Omega} \\rangle\n",
    "\\approx\n",
    "\\sum_{g=1}^P\n",
    "w_g \\, \\omega(\\mathbf{r}_g) \\,\n",
    "n(\\mathbf{r}_g)\n",
    "$$"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "id": "c69310f6-5f8f-417a-9ba5-b879b1915a73",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "outputs": [],
   "source": [
    "grid_level = scf_drv.grid_level\n",
    "if grid_level is None:\n",
    "    grid_level = vlx.dftutils.get_default_grid_level(scf_drv.xcfun)\n",
    "\n",
    "grid_drv = vlx.veloxchemlib.GridDriver()\n",
    "grid_drv.set_level(grid_level)\n",
    "molgrid = grid_drv.generate(molecule)\n",
    "\n",
    "# grid point weights\n",
    "w_g = molgrid.w_to_numpy()\n",
    "\n",
    "# grid point z-coordinates\n",
    "z_g = molgrid.z_to_numpy()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "id": "48aa1c9f-7bf1-4767-a272-4d73f53e8ee5",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Dipole moment:  -1.44687595\n"
     ]
    }
   ],
   "source": [
    "xc_drv = vlx.XCIntegrator()\n",
    "\n",
    "# generate AOs on the grid points\n",
    "chi_g = xc_drv.compute_gto_values(molecule, basis, molgrid)\n",
    "\n",
    "# determine the density on the grid points\n",
    "G = np.einsum(\"ab,bg->ag\", D, chi_g)\n",
    "n_g = np.einsum(\"ag,ag->g\", chi_g, G)\n",
    "\n",
    "print(f\"Dipole moment: {-1.0 * np.dot(w_g, z_g * n_g) : 12.8f}\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "5fa4d927-c87b-459c-85ea-e2337f5db716",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "source": [
    "The numerical accuracy in this integration is limited and determined by the quality of the grid. We here adopted the default grid level in VeloxChem."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7e52a8e0-a861-4a7c-9140-a4e6eeba08b6",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "source": [
    "## Analytic integration\n",
    "\n",
    "If we combine the expressions for the density and the expectation value, we get\n",
    "\n",
    "\\begin{align*}\n",
    "\\langle \\hat{\\Omega} \\rangle  &=\n",
    "\\sum_{\\alpha,\\beta} D_{\\alpha\\beta}\n",
    "\\langle \\chi_\\alpha | \\hat{\\omega} | \\chi_\\beta \\rangle\n",
    "\\end{align*}\n",
    "\n",
    "In form, this Kohn–Sham DFT expression is identical to the result obtained when considering the expectation value of the operator $\\hat{\\Omega}$ with respect to the many-electron Hartree–Fock wave function. It would be incorrect, however, to think of the Kohn–Sham reference state as the wave function of the system.\n",
    "\n",
    "In practice, this expression is used for the evaluation of first-order properties at the DFT level. The scheme is referred to as *analytic* since the property integrals are determined to double precision on the computer and the only numerical integrations are those performed in the [SCF iterations](https://kthpanor.github.io/echem/docs/elec_struct/dft_scf.html)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "id": "67a8df28-97de-4c48-8b5a-cc3d07d5d8ba",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "outputs": [],
   "source": [
    "dipole_drv = vlx.ElectricDipoleIntegralsDriver()\n",
    "dipole_mats = dipole_drv.compute(molecule, basis)\n",
    "\n",
    "mu_z = -1.0 * dipole_mats.z_to_numpy()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "id": "54abf90f-edbc-440d-8583-33e6fd098cc4",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Dipole moment:  -1.44687507\n"
     ]
    }
   ],
   "source": [
    "print(f\"Dipole moment: {np.einsum('ab,ab->', D, mu_z) : 12.8f}\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b4577713-4eda-4952-9e0f-8b58beede8a5",
   "metadata": {
    "editable": true,
    "slideshow": {
     "slide_type": ""
    },
    "tags": []
   },
   "source": [
    "The dipole moment obtained from numerical and analytic integrations are seen to be in very close agreement. The error in the numerical integration is in the order of $10^{-6}$ a.u."
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.7"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
